"""Test the ImJoy RPC module."""
SIO_PORT = 38283
SIO_SERVER_URL = f"http://127.0.0.1:{SIO_PORT}"


def find_item(items, key, value):
    """Find an item with key or attributes in an object list."""
    filtered = [
        item
        for item in items
        if (item[key] if isinstance(item, dict) else getattr(item, key)) == value
    ]
    if len(filtered) == 0:
        return None

    return filtered[0]



"""Provide common pytest fixtures."""
import os
import subprocess
import sys
import time
import uuid

import pytest
import requests
from requests import RequestException
from . import SIO_PORT


JWT_SECRET = str(uuid.uuid4())
os.environ["JWT_SECRET"] = JWT_SECRET
test_env = os.environ.copy()


@pytest.fixture(name="socketio_server", scope="session")
def socketio_server_fixture():
    """Start server as test fixture and tear down after test."""
    with subprocess.Popen(
        [sys.executable, "-m", "hypha.server", f"--port={SIO_PORT}"],
        env=test_env,
    ) as proc:

        timeout = 10
        while timeout > 0:
            try:
                response = requests.get(f"http://127.0.0.1:{SIO_PORT}/health/liveness")
                if response.ok:
                    break
            except RequestException:
                pass
            timeout -= 0.1
            time.sleep(0.1)
        yield
        proc.kill()
        proc.terminate()



"""Test the hypha server."""
import pytest
from imjoy_rpc import connect_to_server
from . import SIO_SERVER_URL
import numpy as np

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class ImJoyPlugin:
    """Represent a test plugin."""

    def __init__(self, ws):
        """Initialize the plugin."""
        self._ws = ws

    async def setup(self):
        """Set up the plugin."""
        await self._ws.log("initialized")

    async def run(self, ctx):
        """Run the plugin."""
        await self._ws.log("hello world")

    async def add(self, data):
        """Add function."""
        return data + 1.0


async def test_connect_to_server(socketio_server):
    """Test connecting to the server."""
    # test workspace is an exception, so it can pass directly
    ws = await connect_to_server(
        {"name": "my plugin", "workspace": "public", "server_url": SIO_SERVER_URL}
    )
    with pytest.raises(Exception, match=r".*Workspace test does not exist.*"):
        ws = await connect_to_server(
            {"name": "my plugin", "workspace": "test", "server_url": SIO_SERVER_URL}
        )
    ws = await connect_to_server({"name": "my plugin", "server_url": SIO_SERVER_URL})
    await ws.export(ImJoyPlugin(ws))

    ws = await connect_to_server({"server_url": SIO_SERVER_URL})
    assert len(ws.config.name) == 36


async def test_numpy_array(socketio_server):
    """Test numpy array."""
    ws = await connect_to_server(
        {"name": "test-plugin", "workspace": "public", "server_url": SIO_SERVER_URL}
    )
    await ws.export(ImJoyPlugin(ws))

    api = await connect_to_server(
        {"name": "client", "workspace": "public", "server_url": SIO_SERVER_URL}
    )
    plugin = await api.get_plugin("test-plugin")
    result = await plugin.add(2.1)
    assert result == 2.1 + 1.0

    large_array = np.zeros([2048, 2048, 4], dtype="float32")
    result = await plugin.add(large_array)
    np.testing.assert_array_equal(result, large_array + 1.0)



"""Provide the RPC."""
import asyncio
import inspect
import io
import logging
import os
import sys
import threading
import traceback
import uuid
import weakref
from collections import OrderedDict
from functools import reduce

from .utils import (
    FuturePromise,
    MessageEmitter,
    ReferenceStore,
    dotdict,
    format_traceback,
)

API_VERSION = "0.2.3"
ALLOWED_MAGIC_METHODS = ["__enter__", "__exit__"]
IO_METHODS = [
    "fileno",
    "seek",
    "truncate",
    "detach",
    "write",
    "read",
    "read1",
    "readall",
    "close",
    "closed",
    "__enter__",
    "__exit__",
    "flush",
    "isatty",
    "__iter__",
    "__next__",
    "readable",
    "readline",
    "readlines",
    "seekable",
    "tell",
    "writable",
    "writelines",
]

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("RPC")


def index_object(obj, ids):
    """Index an object."""
    if isinstance(ids, str):
        return index_object(obj, ids.split("."))
    elif len(ids) == 0:
        return obj
    else:
        if isinstance(obj, dict):
            _obj = obj[ids[0]]
        elif isinstance(obj, (list, tuple)):
            _obj = obj[int(ids[0])]
        else:
            _obj = getattr(obj, ids[0])
        return index_object(_obj, ids[1:])


class RPC(MessageEmitter):
    """Represent the RPC."""

    def __init__(self, connection, rpc_context, config=None, codecs=None):
        """Set up instance."""
        self.manager_api = {}
        self.services = {}
        self._object_store = {}
        self._method_weakmap = weakref.WeakKeyDictionary()
        self._object_weakmap = weakref.WeakKeyDictionary()
        self._local_api = None
        self._remote_set = False
        self._store = ReferenceStore()
        self._remote_interface = None
        self._codecs = codecs or {}
        self.work_dir = os.getcwd()
        self.abort = threading.Event()
        self.id = None

        self.rpc_context = rpc_context

        if config is None:
            config = {}
        self.set_config(config)

        self._remote_logger = dotdict({"info": self._log, "error": self._error})
        super().__init__(self._remote_logger)

        try:
            # FIXME: What exception do we expect?
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        if connection is not None:
            self._connection = connection
            self._setup_handlers(connection)

        self.check_modules()

    def init(self):
        """Initialize the RPC."""
        logger.info("%s initialized", self.config.name)
        self._connection.emit(
            {
                "type": "initialized",
                "config": dict(self.config),
                "peer_id": self._connection.peer_id,
            }
        )

    def reset(self):
        """Reset."""
        self._event_handlers = {}
        self.services = {}
        self._object_store = {}
        self._method_weakmap = weakref.WeakKeyDictionary()
        self._object_weakmap = weakref.WeakKeyDictionary()
        self._local_api = None
        self._remote_set = False
        self._store = ReferenceStore()
        self._remote_interface = None

    def disconnect(self):
        """Disconnect."""
        self._connection.emit({"type": "disconnect"})
        self.reset()
        self._connection.disconnect()

    def default_exit(self):
        """Exit default."""
        logger.info("Terminating plugin: %s", self.id)
        self.abort.set()

    def set_config(self, config):
        """Set config."""
        if config is not None:
            config = dotdict(config)
        else:
            config = dotdict()
        self.id = config.id or self.id or str(uuid.uuid4())
        self.allow_execution = config.allow_execution or False
        self.config = dotdict(
            {
                "allow_execution": self.allow_execution,
                "api_version": API_VERSION,
                "dedicated_thread": True,
                "description": config.description or "[TODO]",
                "id": self.id,
                "lang": "python",
                "name": config.name or "ImJoy RPC Python",
                "type": "rpc-worker",
                "work_dir": self.work_dir,
                "version": config.version or "0.1.0",
            }
        )

    def get_remote(self):
        """Return the remote interface."""
        return self._remote_interface

    def set_interface(self, api, config=None):
        """Set interface."""
        # TODO: setup forwarding_functions
        if config:
            self.set_config(config)

        # convert and store it in a docdict
        # such that the methods are hashable
        if isinstance(api, dict):
            self._local_api = dotdict(
                {
                    a: api[a]
                    for a in api.keys()
                    if not a.startswith("_")
                    or a in ALLOWED_MAGIC_METHODS
                    or a == "_rintf"
                }
            )
        elif inspect.isclass(type(api)):
            self._local_api = dotdict(
                {
                    a: getattr(api, a)
                    for a in dir(api)
                    if not a.startswith("_")
                    or a in ALLOWED_MAGIC_METHODS
                    or a == "_rintf"
                }
            )
        else:
            raise Exception("Invalid api export")

        if not self._remote_set:
            self._fire("interfaceAvailable")
        else:
            self.send_interface()

        # we might installed modules when solving requirements
        # so let's check it again
        self.check_modules()

        fut = self.loop.create_future()

        def done(result):
            if not fut.done():
                fut.set_result(result)

        self.once("interfaceSetAsRemote", done)
        return fut

    def check_modules(self):
        """Check if all the modules exists."""
        try:
            import numpy as np

            self.NUMPY_MODULE = np
        except ImportError:
            self.NUMPY_MODULE = False
            logger.warning(
                "Failed to import numpy, ndarray encoding/decoding will not work"
            )

    def request_remote(self):
        """Request remote interface."""
        self._connection.emit({"type": "getInterface"})

    def send_interface(self):
        """Send interface."""
        if self._local_api is None:
            raise Exception("interface is not set.")

        api = self._encode(self._local_api, True)
        self._connection.emit({"type": "setInterface", "api": api})

    def _dispose_object(self, object_id):
        if object_id in self._object_store:
            del self._object_store[object_id]
        else:
            raise Exception("Object (id={}) not found.".format(object_id))

    def dispose_object(self, obj):
        """Dispose object."""
        if obj in self._object_weakmap:
            object_id = self._object_weakmap[obj]
        else:
            raise Exception("Invalid object")

        def pfunc(resolve, reject):
            """Handle plugin function."""

            def handle_disposed(data):
                """Handle disposed."""
                if "error" in data:
                    reject(data["error"])
                else:
                    resolve(None)

            self._connection.once("disposed", handle_disposed)
            self._connection.emit({"type": "disposeObject", "object_id": object_id})

        return FuturePromise(pfunc, self._remote_logger)

    def _gen_remote_method(self, target_id, name, plugin_id=None):
        """Return remote method."""

        def remote_method(*arguments, **kwargs):
            """Run remote method."""
            arguments = list(arguments)
            # wrap keywords to a dictionary and pass to the last argument
            if kwargs:
                arguments = arguments + [kwargs]

            def pfunc(resolve, reject):
                encoded_promise = self.wrap([resolve, reject])
                # store the key id for removing them from the reference store together
                resolve.__promise_pair = encoded_promise[0]["_rvalue"]
                reject.__promise_pair = encoded_promise[1]["_rvalue"]

                if name in [
                    "register",
                    "registerService",
                    "register_service",
                    "export",
                    "on",
                ]:
                    args = self.wrap(arguments, as_interface=True)
                else:
                    args = self.wrap(arguments)

                call_func = {
                    "type": "method",
                    "target_id": target_id,
                    "name": name,
                    "object_id": plugin_id,
                    "args": args,
                    "with_kwargs": bool(kwargs),
                    "promise": encoded_promise,
                }
                self._connection.emit(call_func)

            return FuturePromise(pfunc, self._remote_logger, self.dispose_object)

        remote_method.__remote_method = True  # pylint: disable=protected-access
        return remote_method

    def _gen_remote_callback(self, target_id, cid, with_promise):
        """Return remote callback."""
        if with_promise:

            def remote_callback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the last argument
                arguments = list(arguments)
                if kwargs:
                    arguments = arguments + [kwargs]

                def pfunc(resolve, reject):
                    encoded_promise = self.wrap([resolve, reject])
                    # store the key id
                    # for removing them from the reference store together
                    resolve.__promise_pair = encoded_promise[0]["_rvalue"]
                    reject.__promise_pair = encoded_promise[1]["_rvalue"]
                    self._connection.emit(
                        {
                            "type": "callback",
                            "id": cid,
                            "target_id": target_id,
                            # 'object_id'  : self.id,
                            "args": self.wrap(arguments),
                            "with_kwargs": bool(kwargs),
                            "promise": encoded_promise,
                        }
                    )

                return FuturePromise(pfunc, self._remote_logger, self.dispose_object)

        else:

            def remote_callback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the last argument
                arguments = list(arguments)
                if kwargs:
                    arguments = arguments + [kwargs]
                self._connection.emit(
                    {
                        "type": "callback",
                        "id": cid,
                        "target_id": target_id,
                        # 'object_id'  : self.id,
                        "args": self.wrap(arguments),
                        "with_kwargs": bool(kwargs),
                    }
                )

        return remote_callback

    def set_remote_interface(self, api):
        """Set remote interface."""
        _remote = self._decode(api, False)
        # update existing interface instead of recreating it
        # this will preserve the object reference
        if self._remote_interface:
            self._remote_interface.clear()
            for k in _remote:
                self._remote_interface[k] = _remote[k]
        else:
            self._remote_interface = _remote
        self._fire("remoteReady")
        self._run_with_context(self._set_remote_api, _remote)

    def _set_remote_api(self, _remote):
        """Set remote API."""
        self.rpc_context.api = _remote
        self.rpc_context.api.WORK_DIR = self.work_dir
        if "config" not in self.rpc_context.api:
            self.rpc_context.api.config = dotdict()
        self.rpc_context.api.config.work_dir = self.work_dir

    def _log(self, info):
        self._connection.emit({"type": "log", "message": info})

    def _error(self, error):
        self._connection.emit({"type": "error", "message": error})

    def _call_method(
        self, method, args, kwargs, resolve=None, reject=None, method_name=None
    ):
        try:
            result = method(*args, **kwargs)
            if result is not None and inspect.isawaitable(result):

                async def _wait(result):
                    try:
                        result = await result
                        if resolve is not None:
                            resolve(result)
                        elif result is not None:
                            logger.debug("returned value %s", result)
                    except Exception as err:
                        traceback_error = traceback.format_exc()
                        logger.exception("Error in method %s", err)
                        self._connection.emit(
                            {"type": "error", "message": traceback_error}
                        )
                        if reject is not None:
                            reject(Exception(format_traceback(traceback_error)))

                asyncio.ensure_future(_wait(result))
            else:
                if resolve is not None:
                    resolve(result)
        except Exception as err:
            traceback_error = traceback.format_exc()
            logger.error("Error in method %s: %s", method_name, err)
            self._connection.emit({"type": "error", "message": traceback_error})
            if reject is not None:
                reject(Exception(format_traceback(traceback_error)))

    def _run_with_context(self, func, *args, **kwargs):
        self.rpc_context.run_with_context(self.id, func, *args, **kwargs)

    def _setup_handlers(self, connection):
        connection.on("init", self.init)
        connection.on("execute", self._handle_execute)
        connection.on("method", self._handle_method)
        connection.on("callback", self._handle_callback)
        connection.on("error", self._handle_error)
        connection.on("disconnected", self._disconnected_hanlder)
        connection.on("getInterface", self._get_interface_handler)
        connection.on("setInterface", self._set_interface_handler)
        connection.on("interfaceSetAsRemote", self._remote_set_handler)
        connection.on("disposeObject", self._dispose_object_handler)

    def _dispose_object_handler(self, data):
        try:
            self._dispose_object(data["object_id"])
            self._connection.emit({"type": "disposed"})
        except Exception as e:
            logger.error("failed to dispose object: %s", e)
            self._connection.emit({"type": "disposed", "error": str(e)})

    def _disconnected_hanlder(self, data):
        self._fire("beforeDisconnect")
        self._connection.disconnect()
        self._fire("disconnected", data)

    def _get_interface_handler(self, data):
        if self._local_api is not None:
            self.send_interface()
        else:
            self.once("interfaceAvailable", self.send_interface)

    def _set_interface_handler(self, data):
        self.set_remote_interface(data["api"])
        self._connection.emit({"type": "interfaceSetAsRemote"})

    def _remote_set_handler(self, data):
        self._remote_set = True
        self._fire("interfaceSetAsRemote")

    def _handle_execute(self, data):
        if self.allow_execution:
            try:
                t = data["code"]["type"]
                if t == "script":
                    content = data["code"]["content"]
                    # TODO: fix the imjoy module such that it will
                    # stick to the current context api
                    exec(content)
                elif t == "requirements":
                    pass
                else:
                    raise Exception("unsupported type")
                self._connection.emit({"type": "executed"})
            except Exception as err:
                traceback_error = traceback.format_exc()
                logger.exception("Error during execution: %s", err)
                self._connection.emit({"type": "executed", "error": traceback_error})
        else:
            self._connection.emit(
                {"type": "executed", "error": "execution is not allowed"}
            )
            logger.warn("execution is blocked due to allow_execution=False")

    def _handle_method(self, data):
        reject = None
        try:
            if "promise" in data:
                resolve, reject = self.unwrap(data["promise"], False)
            _interface = self._object_store[data["object_id"]]
            method = index_object(_interface, data["name"])
            if "promise" in data:
                args = self.unwrap(data["args"], True)
                if data.get("with_kwargs"):
                    kwargs = args.pop()
                else:
                    kwargs = {}
                # args.append({'id': self.id})
                self._run_with_context(
                    self._call_method,
                    method,
                    args,
                    kwargs,
                    resolve=resolve,
                    reject=reject,
                    method_name=data["name"],
                )
            else:
                args = self.unwrap(data["args"], True)
                if data.get("with_kwargs"):
                    kwargs = args.pop()
                else:
                    kwargs = {}
                # args.append({'id': self.id})
                self._run_with_context(
                    self._call_method, method, args, kwargs, method_name=data["name"]
                )
        except Exception as err:
            traceback_error = traceback.format_exc()
            logger.exception("Error during calling method: %s", err)
            self._connection.emit({"type": "error", "message": traceback_error})
            if callable(reject):
                reject(traceback_error)

    def _handle_callback(self, data):
        reject = None
        try:
            if "promise" in data:
                resolve, reject = self.unwrap(data["promise"], False)
                method = self._store.fetch(data["id"])
                if method is None:
                    raise Exception(
                        "Callback function can only called once, "
                        "if you want to call a function for multiple times, "
                        "please make it as a plugin api function. "
                        "See https://imjoy.io/docs for more details."
                    )
                args = self.unwrap(data["args"], True)
                if data.get("with_kwargs"):
                    kwargs = args.pop()
                else:
                    kwargs = {}
                self._run_with_context(
                    self._call_method,
                    method,
                    args,
                    kwargs,
                    resolve=resolve,
                    reject=reject,
                    method_name=data["id"],
                )

            else:
                method = self._store.fetch(data["id"])
                if method is None:
                    raise Exception(
                        "Callback function can only called once, "
                        "if you want to call a function for multiple times, "
                        "please make it as a plugin api function. "
                        "See https://imjoy.io/docs for more details."
                    )
                args = self.unwrap(data["args"], True)
                if data.get("with_kwargs"):
                    kwargs = args.pop()
                else:
                    kwargs = {}
                self._run_with_context(
                    self._call_method, method, args, kwargs, method_name=data["id"]
                )
        except Exception as err:
            traceback_error = traceback.format_exc()
            logger.exception("error when calling callback function: %s", err)
            self._connection.emit({"type": "error", "message": traceback_error})
            if callable(reject):
                reject(traceback_error)

    def _handle_error(self, detail):
        self._fire("error", detail)

    def wrap(self, args, as_interface=False):
        """Wrap arguments."""
        wrapped = self._encode(args, as_interface=as_interface)
        return wrapped

    def _encode(self, a_object, as_interface=False, object_id=None):
        """Encode object."""
        if isinstance(a_object, (int, float, bool, str, bytes)) or a_object is None:
            return a_object

        if callable(a_object):
            if as_interface:
                if not object_id:
                    raise Exception("object_id is not specified.")
                b_object = {
                    "_rtype": "interface",
                    "_rtarget_id": self._connection.peer_id,
                    "_rintf": object_id,
                    "_rvalue": as_interface,
                }
                try:
                    self._method_weakmap[a_object] = b_object
                except Exception:
                    pass
            elif a_object in self._method_weakmap:
                b_object = self._method_weakmap[a_object]
            else:
                cid = self._store.put(a_object)
                b_object = {
                    "_rtype": "callback",
                    # Some functions do not have the __name__ attribute
                    # for example when we use functools.partial to create functions
                    "_rname": getattr(a_object, "__name__", cid),
                    "_rtarget_id": self._connection.peer_id,
                    "_rvalue": cid,
                }
            return b_object

        if isinstance(a_object, tuple):
            a_object = list(a_object)

        if isinstance(a_object, dict):
            a_object = dict(a_object)

        # skip if already encoded
        if isinstance(a_object, dict) and "_rtype" in a_object:
            # make sure the interface functions are encoded
            if "_rintf" in a_object:
                temp = a_object["_rtype"]
                del a_object["_rtype"]
                b_object = self._encode(a_object, as_interface, object_id)
                b_object._rtype = temp
            else:
                b_object = a_object
            return b_object

        isarray = isinstance(a_object, list)
        b_object = None

        encoded_obj = None
        for tp in self._codecs:
            codec = self._codecs[tp]
            if codec.encoder and isinstance(a_object, codec.type):
                # TODO: what if multiple encoders found
                encoded_obj = codec.encoder(a_object)
                if isinstance(encoded_obj, dict) and "_rtype" not in encoded_obj:
                    encoded_obj["_rtype"] = codec.name
                # encode the functions in the interface object
                if isinstance(encoded_obj, dict) and "_rintf" in encoded_obj:
                    temp = encoded_obj["_rtype"]
                    del encoded_obj["_rtype"]
                    encoded_obj = self._encode(encoded_obj, True)
                    encoded_obj["_rtype"] = temp
                b_object = encoded_obj
                return b_object

        if self.NUMPY_MODULE and isinstance(
            a_object, (self.NUMPY_MODULE.ndarray, self.NUMPY_MODULE.generic)
        ):
            v_bytes = a_object.tobytes()
            b_object = {
                "_rtype": "ndarray",
                "_rvalue": v_bytes,
                "_rshape": a_object.shape,
                "_rdtype": str(a_object.dtype),
            }

        elif isinstance(a_object, Exception):
            b_object = {"_rtype": "error", "_rvalue": str(a_object)}
        elif isinstance(a_object, memoryview):
            b_object = {"_rtype": "memoryview", "_rvalue": a_object.tobytes()}
        elif isinstance(
            a_object, (io.IOBase, io.TextIOBase, io.BufferedIOBase, io.RawIOBase)
        ):
            b_object = {
                m: getattr(a_object, m) for m in IO_METHODS if hasattr(a_object, m)
            }
            b_object["_rintf"] = True
            b_object = self._encode(b_object)

        # NOTE: "typedarray" is not used
        elif isinstance(a_object, OrderedDict):
            b_object = {
                "_rtype": "orderedmap",
                "_rvalue": self._encode(list(a_object), as_interface),
            }
        elif isinstance(a_object, set):
            b_object = {
                "_rtype": "set",
                "_rvalue": self._encode(list(a_object), as_interface),
            }
        elif hasattr(a_object, "_rintf") and a_object._rintf is True:
            b_object = self._encode(a_object, True)
        elif isinstance(a_object, (list, dict)) or inspect.isclass(type(a_object)):
            b_object = [] if isarray else {}
            if not isinstance(a_object, (list, dict)) and inspect.isclass(
                type(a_object)
            ):
                a_object_norm = {
                    a: getattr(a_object, a)
                    for a in dir(a_object)
                    if not a.startswith("_") or a in ALLOWED_MAGIC_METHODS
                }
                # always encode class instance as interface
                as_interface = True
            else:
                a_object_norm = a_object

            keys = range(len(a_object_norm)) if isarray else a_object_norm.keys()
            # encode interfaces
            if (not isarray and a_object_norm.get("_rintf")) or as_interface:
                if object_id is None:
                    _rintf = a_object_norm.get("_rintf") if not isarray else None
                    if _rintf and isinstance(_rintf, str):
                        object_id = _rintf  # enable custom object_id
                    else:
                        object_id = str(uuid.uuid4())
                    # Note: object with the same id will be overwritten
                    if object_id in self._object_store:
                        logger.warn(
                            "Overwritting interface object with the same id: %s",
                            object_id,
                        )
                    self._object_store[object_id] = a_object

                has_function = False
                for key in keys:
                    if isinstance(key, str) and (
                        key.startswith("_") and key not in ALLOWED_MAGIC_METHODS
                    ):
                        continue
                    encoded = self._encode(
                        a_object_norm[key],
                        as_interface + "." + str(key)
                        if isinstance(as_interface, str)
                        else str(key),
                        # We need to convert to a string here,
                        # otherwise 0 will not be truthy.
                        object_id,
                    )
                    if callable(a_object_norm[key]):
                        has_function = True
                    if isarray:
                        b_object.append(encoded)
                    else:
                        b_object[key] = encoded
                # TODO: how to despose list object? create a wrapper for list?
                if not isarray and has_function:
                    b_object["_rintf"] = object_id
                # remove interface when closed
                if "on" in a_object_norm and callable(a_object_norm["on"]):

                    def remove_interface(_):
                        if object_id in self._object_store:
                            del self._object_store[object_id]

                    a_object_norm["on"]("close", remove_interface)
            else:
                for key in keys:
                    if isarray:
                        b_object.append(self._encode(a_object_norm[key]))
                    else:
                        b_object[key] = self._encode(a_object_norm[key])
        else:
            raise Exception("imjoy-rpc: Unsupported data type:" + str(a_object))
        return b_object

    def unwrap(self, args, with_promise):
        """Unwrap arguments."""
        # wraps each callback so that the only one could be called
        result = self._decode(args, with_promise)
        return result

    def _decode(self, a_object, with_promise):
        """Decode object."""
        if a_object is None:
            return a_object
        if isinstance(a_object, dict) and "_rtype" in a_object:
            b_object = None
            if (
                self._codecs.get(a_object["_rtype"])
                and self._codecs[a_object["_rtype"]].decoder
            ):
                if "_rintf" in a_object:
                    temp = a_object["_rtype"]
                    del a_object["_rtype"]
                    a_object = self._decode(a_object, with_promise)
                    a_object["_rtype"] = temp
                b_object = self._codecs[a_object["_rtype"]].decoder(a_object)
            elif a_object["_rtype"] == "callback":
                b_object = self._gen_remote_callback(
                    a_object.get("_rtarget_id"), a_object["_rvalue"], with_promise
                )
            elif a_object["_rtype"] == "interface":
                b_object = self._gen_remote_method(
                    a_object.get("_rtarget_id"), a_object["_rvalue"], a_object["_rintf"]
                )
            elif a_object["_rtype"] == "ndarray":
                # create build array/tensor if used in the plugin
                try:
                    if isinstance(a_object["_rvalue"], (list, tuple)):
                        a_object["_rvalue"] = reduce(
                            (lambda x, y: x + y), a_object["_rvalue"]
                        )
                    # make sure we have bytes instead of memoryview, e.g. for Pyodide
                    elif isinstance(a_object["_rvalue"], memoryview):
                        a_object["_rvalue"] = a_object["_rvalue"].tobytes()
                    elif not isinstance(a_object["_rvalue"], bytes):
                        raise Exception(
                            "Unsupported data type: " + str(type(a_object["_rvalue"]))
                        )
                    if self.NUMPY_MODULE:
                        b_object = self.NUMPY_MODULE.frombuffer(
                            a_object["_rvalue"], dtype=a_object["_rdtype"]
                        ).reshape(tuple(a_object["_rshape"]))

                    else:
                        b_object = a_object
                        logger.warn("numpy is not available, failed to decode ndarray")

                except Exception as exc:
                    logger.debug("Error in converting: %s", exc)
                    b_object = a_object
                    raise exc
            elif a_object["_rtype"] == "memoryview":
                b_object = memoryview(a_object["_rvalue"])
            elif a_object["_rtype"] == "blob":
                if isinstance(a_object["_rvalue"], str):
                    b_object = io.StringIO(a_object["_rvalue"])
                elif isinstance(a_object["_rvalue"], bytes):
                    b_object = io.BytesIO(a_object["_rvalue"])
                else:
                    raise Exception(
                        "Unsupported blob value type: " + str(type(a_object["_rvalue"]))
                    )
            elif a_object["_rtype"] == "typedarray":
                if self.NUMPY_MODULE:
                    b_object = self.NUMPY_MODULE.frombuffer(
                        a_object["_rvalue"], dtype=a_object["_rdtype"]
                    )
                else:
                    b_object = a_object["_rvalue"]
            elif a_object["_rtype"] == "orderedmap":
                b_object = OrderedDict(self._decode(a_object["_rvalue"], with_promise))
            elif a_object["_rtype"] == "set":
                b_object = set(self._decode(a_object["_rvalue"], with_promise))
            elif a_object["_rtype"] == "error":
                b_object = Exception(a_object["_rvalue"])
            else:
                # make sure all the interface functions are decoded
                if "_rintf" in a_object:
                    temp = a_object["_rtype"]
                    del a_object["_rtype"]
                    a_object = self._decode(a_object, with_promise)
                    a_object["_rtype"] = temp
                b_object = a_object
        elif isinstance(a_object, (dict, list, tuple)):
            if isinstance(a_object, tuple):
                a_object = list(a_object)
            isarray = isinstance(a_object, list)
            b_object = [] if isarray else dotdict()
            keys = range(len(a_object)) if isarray else a_object.keys()
            for key in keys:
                val = a_object[key]
                if isarray:
                    b_object.append(self._decode(val, with_promise))
                else:
                    b_object[key] = self._decode(val, with_promise)
        # make sure we have bytes instead of memoryview, e.g. for Pyodide
        elif isinstance(a_object, memoryview):
            b_object = a_object.tobytes()
        elif isinstance(a_object, bytearray):
            b_object = bytes(a_object)
        else:
            b_object = a_object

        # object id, used for dispose the object
        if isinstance(a_object, dict) and a_object.get("_rintf"):
            # make the dict hashable
            if isinstance(b_object, dict):
                if not isinstance(b_object, dotdict):
                    b_object = dotdict(b_object)
                # __rid__ is used for hashing the object for removing it afterwards
                b_object.__rid__ = a_object.get("_rintf")

            self._object_weakmap[b_object] = a_object.get("_rintf")
        return b_object



"""Provide the connection."""
import asyncio
import gzip
import logging
import math
import sys
import time
import uuid

import msgpack

from imjoy_rpc.utils import MessageEmitter, dotdict

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("core-connection")
logger.setLevel(logging.WARNING)

all_connections = {}
CHUNK_SIZE = 1024 * 1000


def send_as_msgpack(msg, send, accept_encoding):
    """Send the message by using msgpack encoding."""
    encoded = {
        "type": "msgpack_data",
        "msg_type": msg.pop("type"),
    }
    if msg.get("peer_id"):
        encoded["peer_id"] = msg.pop("peer_id")
    if msg.get("plugin_id"):
        encoded["plugin_id"] = msg.pop("plugin_id")

    packed = msgpack.packb(msg, use_bin_type=True)
    total_size = len(packed)
    if total_size > CHUNK_SIZE and "gzip" in accept_encoding:
        compressed = gzip.compress(packed)
        # Only send the compressed version
        # if the compression ratio is > 80%;
        if len(compressed) <= total_size * 0.8:
            packed = compressed
            encoded["compression"] = "gzip"

    total_size = len(packed)
    if total_size <= CHUNK_SIZE:
        encoded["data"] = packed
        asyncio.ensure_future(send(encoded))
    else:

        async def send_chunks():
            object_id = str(uuid.uuid4())
            chunk_num = int(math.ceil(float(total_size) / CHUNK_SIZE))
            # send chunk by chunk
            for idx in range(chunk_num):
                start_byte = idx * CHUNK_SIZE
                chunk = {
                    "type": "msgpack_chunk",
                    "object_id": object_id,
                    "data": packed[start_byte : start_byte + CHUNK_SIZE],
                    "index": idx,
                    "total": chunk_num,
                }
                logger.info(
                    "Sending chunk %d/%d (%d bytes)",
                    idx + 1,
                    chunk_num,
                    total_size,
                )
                await send(chunk)

            # reference the chunked object
            encoded["chunked_object"] = object_id
            await send(encoded)

        asyncio.ensure_future(send_chunks())


def decode_msgpack(data, chunk_store):
    """Try to decode the data as msgpack."""
    dtype = data.get("type")
    if dtype == "msgpack_chunk":
        id_ = data["object_id"]
        if id_ not in chunk_store:
            chunk_store[id_] = []
        assert data["index"] == len(chunk_store[id_])
        chunk_store[id_].append(data["data"])
        return

    if dtype == "msgpack_data":
        if data.get("chunked_object"):
            object_id = data["chunked_object"]
            chunks = chunk_store[object_id]
            del chunk_store[object_id]
            data["data"] = b"".join(chunks)
        if data.get("compression"):
            if data["compression"] == "gzip":
                data["data"] = gzip.decompress(data["data"])
            else:
                raise Exception(f"Unsupported compression: {data['compression']}")
        decoded = msgpack.unpackb(data["data"], use_list=False, raw=False)
        if data.get("plugin_id"):
            decoded["plugin_id"] = data.get("plugin_id")
        if data.get("peer_id"):
            decoded["peer_id"] = data.get("peer_id")
        decoded["type"] = data["msg_type"]
        data = decoded

    return data


class BasicConnection(MessageEmitter):
    """Represent a base connection."""

    def __init__(self, send):
        """Set up instance."""
        super().__init__(logger)
        self.plugin_config = dotdict()
        self._access_token = None
        self._expires_in = None
        self._plugin_origin = "*"
        self._refresh_token = None
        self._send = send
        self.peer_id = None
        self.on("initialized", self._initialized)
        self._chunk_store = {}
        self.accept_encoding = []

    def _initialized(self, data):
        self.plugin_config = data["config"]
        # peer_id can only be set for once
        self.peer_id = data["peer_id"]
        self._plugin_origin = data.get("origin", "*")
        all_connections[self.peer_id] = self
        if self._plugin_origin != "*":
            logger.info(
                "Connection to the imjoy-rpc peer $%s is limited to origin %s.",
                self.peer_id,
                self._plugin_origin,
            )

        if not self.peer_id:
            raise Exception("Please provide a peer_id for the connection.")

        if self.plugin_config.get("auth"):
            if self._plugin_origin == "*":
                logger.error(
                    "Refuse to transmit the token without an explicit origin, "
                    "there is a security risk that you may leak the credential "
                    "to website from other origin. "
                    "Please specify the `origin` explicitly."
                )
                self._access_token = None
                self._refresh_token = None

            if self.plugin_config["auth"]["type"] != "jwt":
                logger.error(
                    "Unsupported authentication type: %s", self.plugin_config.auth.type
                )
            else:
                self._expires_in = self.plugin_config["auth"]["expires_in"]
                self._access_token = self.plugin_config["auth"]["access_token"]
                self._refresh_token = self.plugin_config["auth"]["refresh_token"]

    def handle_message(self, data):
        """Handle a message."""
        data = decode_msgpack(data, self._chunk_store)
        if data is None:
            return

        target_id = data.get("target_id")
        if target_id and self.peer_id and target_id != self.peer_id:
            conn = all_connections[target_id]
            if conn:
                conn._fire(data["type"], data)  # pylint: disable=protected-access
            else:
                logger.warning(
                    "Connection with target_id %s not found, discarding data: %s",
                    target_id,
                    data,
                )
        else:
            if data.get("type") == "initialized":
                self.accept_encoding = data.get("accept_encoding", [])
            self._fire(data["type"], data)

    def connect(self):
        """Connect."""
        self._fire("connected")

    async def execute(self, code):
        """Execute."""
        # pylint: disable=no-self-use
        raise PermissionError

    def emit(self, msg):
        """Send a message to the plugin site."""
        if self._access_token:
            if time.time() >= self._expires_in:
                # TODO: refresh access token
                raise Exception("Refresh token is not implemented.")

            msg["access_token"] = self._access_token
        msg["peer_id"] = msg.get("peer_id") or self.peer_id
        if msg.get("type") in ["initialize"] or "msgpack" not in self.accept_encoding:
            if msg.get("type") == "initialize":
                # Notify the plugin that the server supports msgpack decoding
                msg["accept_encoding"] = ["msgpack", "gzip"]
            asyncio.ensure_future(self._send(msg))
        else:
            send_as_msgpack(msg, self._send, self.accept_encoding)

    def disconnect(self):
        """Disconnect the plugin."""
        self.emit({"type": "disconnect"})
        if self.peer_id and self.peer_id in all_connections:
            del all_connections[self.peer_id]



"""Provide utility functions for RPC."""
import sys

if sys.version_info < (3, 7):
    import aiocontextvars  # noqa: F401

import asyncio
import contextvars
import copy
import io
import locale
import os
import secrets
import string
import threading
import traceback
import uuid

from .werkzeug.local import Local


def generate_password(length=50):
    """Generate a password."""
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for i in range(length))


_hash_id = generate_password()


class dotdict(dict):  # pylint: disable=invalid-name
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __setattr__(self, name, value):
        """Set the attribute."""
        # Make an exception for __rid__
        if name == "__rid__":
            super().__setattr__("__rid__", value)
        else:
            super().__setitem__(name, value)

    def __hash__(self):
        """Return the hash."""
        if self.__rid__ and type(self.__rid__) is str:
            return hash(self.__rid__ + _hash_id)

        # FIXME: This does not address the issue of inner list
        return hash(tuple(sorted(self.items())))

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))


def format_traceback(traceback_string):
    """Format traceback."""
    formatted_lines = traceback_string.splitlines()
    # remove the second and third line
    formatted_lines.pop(1)
    formatted_lines.pop(1)
    formatted_error_string = "\n".join(formatted_lines)
    formatted_error_string = formatted_error_string.replace(
        'File "<string>"', "Plugin script"
    )
    return formatted_error_string


class ReferenceStore:
    """Represent a reference store."""

    def __init__(self):
        """Set up store."""
        self._store = {}

    @staticmethod
    def _gen_id():
        """Generate an id."""
        return str(uuid.uuid4())

    def put(self, obj):
        """Put an object into the store."""
        id_ = self._gen_id()
        self._store[id_] = obj
        return id_

    def fetch(self, search_id):
        """Fetch an object from the store by id."""
        if search_id not in self._store:
            return None
        obj = self._store[search_id]
        if not hasattr(obj, "__remote_method"):
            del self._store[search_id]
        if hasattr(obj, "__promise_pair"):
            self.fetch(obj.__promise_pair)
        return obj


class Promise(object):  # pylint: disable=useless-object-inheritance
    """Represent a promise."""

    def __init__(self, pfunc, logger=None):
        """Set up promise."""
        self._resolve_handler = None
        self._finally_handler = None
        self._catch_handler = None
        self._logger = logger

        def resolve(*args, **kwargs):
            self.resolve(*args, **kwargs)

        def reject(*args, **kwargs):
            self.reject(*args, **kwargs)

        pfunc(resolve, reject)

    def resolve(self, result):
        """Resolve promise."""
        try:
            if self._resolve_handler:
                self._resolve_handler(result)
        except Exception as exc:  # pylint: disable=broad-except
            if self._catch_handler:
                self._catch_handler(exc)
            elif not self._finally_handler:
                if self._logger:
                    self._logger.error("Uncaught Exception: {}".format(exc))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def reject(self, error):
        """Reject promise."""
        try:
            if self._catch_handler:
                self._catch_handler(error)
            elif not self._finally_handler:
                if self._logger:
                    self._logger.error("Uncaught Exception: {}".format(error))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def then(self, handler):
        """Implement then callback.

        Set handler and return the promise.
        """
        self._resolve_handler = handler
        return self

    def finally_(self, handler):
        """Implement finally callback.

        Set handler and return the promise.
        """
        self._finally_handler = handler
        return self

    def catch(self, handler):
        """Implement catch callback.

        Set handler and return the promise.
        """
        self._catch_handler = handler
        return self


class FuturePromise(Promise, asyncio.Future):
    """Represent a promise as a future."""

    def __init__(self, pfunc, logger=None, dispose=None):
        """Set up promise."""
        Promise.__init__(self, pfunc, logger)
        asyncio.Future.__init__(self)
        self.__dispose = dispose
        self.__obj = None

    async def __aenter__(self):
        """Enter context for async."""
        ret = await self
        if isinstance(ret, dict):
            if "__enter__" in ret:
                ret = await ret["__enter__"]()
            self.__obj = ret
        return ret

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Exit context for async."""
        if self.__obj:
            if "__exit__" in self.__obj:
                await self.__obj["__exit__"]()
            if self.__dispose:
                await self.__dispose(self.__obj)
            del self.__obj

    def resolve(self, result):
        """Resolve promise."""
        if self._resolve_handler or self._finally_handler:
            super().resolve(result)
        else:
            self.set_result(result)

    def reject(self, error):
        """Reject promise."""
        if self._catch_handler or self._finally_handler:
            super().reject(error)
        else:
            if error:
                self.set_exception(Exception(str(error)))
            else:
                self.set_exception(Exception())


class MessageEmitter:
    """Represent a message emitter."""

    def __init__(self, logger=None):
        """Set up instance."""
        self._event_handlers = {}
        self._logger = logger

    def on(self, event, handler):
        """Register an event handler."""
        if event not in self._event_handlers:
            self._event_handlers[event] = []
        self._event_handlers[event].append(handler)

    def once(self, event, handler):
        """Register an event handler that should only run once."""
        # wrap the handler function,
        # this is needed because setting property
        # won't work for member function of a class instance
        def wrap_func(*args, **kwargs):
            return handler(*args, **kwargs)

        wrap_func.___event_run_once = True
        self.on(event, wrap_func)

    def off(self, event=None, handler=None):
        """Reset one or all event handlers."""
        if event is None and handler is None:
            self._event_handlers = {}
        elif event is not None and handler is None:
            if event in self._event_handlers:
                self._event_handlers[event] = []
        else:
            if event in self._event_handlers:
                self._event_handlers[event].remove(handler)

    def emit(self, msg):
        """Emit a message."""
        raise NotImplementedError

    def _fire(self, event, data=None):
        """Fire an event handler."""
        if event in self._event_handlers:
            for handler in self._event_handlers[event]:
                try:
                    handler(data)
                except Exception as err:
                    traceback_error = traceback.format_exc()
                    if self._logger:
                        self._logger.exception(err)
                    self.emit({"type": "error", "message": traceback_error})
                finally:
                    if hasattr(handler, "___event_run_once"):
                        self._event_handlers[event].remove(handler)
        else:
            if self._logger and self._logger.debug:
                self._logger.debug("Unhandled event: {}, data: {}".format(event, data))


class ContextLocal(Local):
    """Represent a local context."""

    def __init__(self, default_context_id=None):
        """Set up instance."""
        if default_context_id is None:
            default_context_id = "_"
        object.__setattr__(
            self, "__context_id__", contextvars.ContextVar("context_id", default=None)
        )
        object.__setattr__(self, "__thread_lock__", threading.Lock())
        object.__setattr__(self, "__storage__", {})
        object.__setattr__(self, "__ident_func__", self.__get_ident)
        object.__setattr__(self, "__default_context_id__", default_context_id)

    def set_default_context(self, context_id):
        """Set the default context."""
        object.__setattr__(self, "__default_context_id__", context_id)

    def run_with_context(self, context_id, func, *args, **kwargs):
        """Run with the context."""
        # make sure we obtain the context with the correct context_id
        with object.__getattribute__(self, "__thread_lock__"):
            object.__getattribute__(self, "__context_id__").set(context_id)
            object.__setattr__(self, "__default_context_id__", context_id)
            ctx = contextvars.copy_context()
        ctx.run(func, *args, **kwargs)

    def __get_ident(self):
        """Return the context identity."""
        ident = object.__getattribute__(self, "__context_id__").get()
        if ident is None:
            return object.__getattribute__(self, "__default_context_id__")
        if not isinstance(ident, str):
            raise ValueError("Context value must be a string.")
        return ident


def encode_zarr_store(zobj):
    """Encode the zarr store."""
    import zarr

    path_prefix = f"{zobj.path}/" if zobj.path else ""

    def getItem(key, options=None):
        return zobj.store[path_prefix + key]

    def setItem(key, value):
        zobj.store[path_prefix + key] = value

    def containsItem(key, options=None):
        if path_prefix + key in zobj.store:
            return True

    return {
        "_rintf": True,
        "_rtype": "zarr-array" if isinstance(zobj, zarr.Array) else "zarr-group",
        "getItem": getItem,
        "setItem": setItem,
        "containsItem": containsItem,
    }


def register_default_codecs(options=None):
    """Register default codecs."""
    from imjoy_rpc import api

    if options is None or "zarr-array" in options:
        import zarr

        api.registerCodec(
            {"name": "zarr-array", "type": zarr.Array, "encoder": encode_zarr_store}
        )

    if options is None or "zarr-group" in options:
        import zarr

        api.registerCodec(
            {"name": "zarr-group", "type": zarr.Group, "encoder": encode_zarr_store}
        )


def setup_js_socketio(config, resolve, reject):
    """Load socketio in javascript."""
    from js import eval

    script = """
    let loadScript;
    if (
        typeof WorkerGlobalScope !== "undefined" &&
        self instanceof WorkerGlobalScope
    ) {
        loadScript = async function(url) {
            importScripts(url);
        }
    }
    else{
        function _importScript(url) {
            //url is URL of external file, implementationCode is the code
            //to be called from the file, location is the location to
            //insert the <script> element
            return new Promise((resolve, reject) => {
            var scriptTag = document.createElement("script");
            scriptTag.src = url;
            scriptTag.type = "text/javascript";
            scriptTag.onload = resolve;
            scriptTag.onreadystatechange = function() {
                if (this.readyState === "loaded" || this.readyState === "complete") {
                resolve();
                }
            };
            scriptTag.onerror = reject;
            document.head.appendChild(scriptTag);
            });
        }

        // support loadScript outside web worker
        loadScript = async function() {
            var args = Array.prototype.slice.call(arguments),
            len = args.length,
            i = 0;
            for (; i < len; i++) {
            await _importScript(args[i]);
            }
        }
    }
    const socketio_lib_url = "https://cdn.jsdelivr.net/npm/" +
    "socket.io-client@4.0.1/dist/socket.io.min.js";
    globalThis.setupJSSocketIo = async function(config, resolve, reject){
        try{
            await loadScript(socketio_lib_url)
            const toObject = (x) => {
                if(x===undefined || x===null) return x;
                return x.toJs({dict_converter : Object.fromEntries})
            }
            config = toObject(config)
            config.server_token=config.token
            globalThis.config = config
            const url = config.server_url;
            const extraHeaders = {};
            if (config.token) {
                extraHeaders.Authorization = "Bearer " + config.token;
            }
            // const basePath = new URL(url).pathname;
            // Note: extraHeaders only works for polling transport (the default)
            // If we switch to websocket only, the headers won't be respected
            if(globalThis.socket){
                globalThis.socket.disconnect();
            }
            const socket = io(url, {
                withCredentials: true,
                extraHeaders,
            });
            let connected = false;
            socket.on("connect", () => {
                if(connected) {
                    console.warn("Skipping reconnect to the server");
                    return
                }

                globalThis.sendMessage = function(data){

                    return new Promise((resolve, reject)=>{
                        try{
                            data = toObject(data)
                            socket.emit("plugin_message", data, result => {
                                if (!result.success) reject(result.detail)
                                else resolve(result)
                            })
                        }
                        catch(e){
                            reject(e)
                        }

                    })
                }

                socket.emit("register_plugin", config, async (result) => {
                    if (!result.success) {
                        console.error(result.detail);
                        reject(result.detail);
                        return;
                    }
                    connected = true;
                    globalThis.setMessageCallback = (cb)=>{
                        socket.on("plugin_message", cb);
                    }
                    console.log("Plugin registered: " + config.name)
                    resolve(config.plugin_id);
                })

                socket.on("connect_error", (error) => {
                    console.error("connection error", error);
                    reject(`${error}`);
                });
                socket.on("disconnect", () => {
                    console.error("disconnected");
                    reject("disconnected");
                });
            })
            globalThis.socket = socket;
        }
        catch(e){
            reject(`${error}`);
        }
    }
    """

    eval(script)(config, resolve, reject)


def setup_connection(
    _rpc_context,
    connection_type,
    logger=None,
    on_ready_callback=None,
    on_error_callback=None,
):
    """Set up the connection."""
    if connection_type == "jupyter":
        if logger:
            logger.info("Using jupyter connection for imjoy-rpc")
        from .connection.jupyter_connection import JupyterCommManager

        manager = JupyterCommManager(_rpc_context)
        _rpc_context.api = dotdict(
            init=manager.init,
            export=manager.set_interface,
            registerCodec=manager.register_codec,
            register_codec=manager.register_codec,
        )
        manager.start(
            on_ready_callback=on_ready_callback, on_error_callback=on_error_callback
        )
    elif connection_type == "colab":
        if logger:
            logger.info("Using colab connection for imjoy-rpc")
        from .connection.colab_connection import ColabManager

        manager = ColabManager(_rpc_context)
        _rpc_context.api = dotdict(
            init=manager.init,
            export=manager.set_interface,
            registerCodec=manager.register_codec,
            register_codec=manager.register_codec,
        )
        manager.start(
            on_ready_callback=on_ready_callback, on_error_callback=on_error_callback
        )
    elif connection_type == "terminal":
        if logger:
            logger.info("Using socketio connection for imjoy-rpc")
        from .connection.socketio_connection import SocketIOManager

        manager = SocketIOManager(_rpc_context)
        _rpc_context.api = dotdict(
            init=manager.init,
            export=manager.set_interface,
            registerCodec=manager.register_codec,
            register_codec=manager.register_codec,
        )
        manager.start(
            _rpc_context.default_config.get("server_url"),
            _rpc_context.default_config.get("token"),
            on_ready_callback=on_ready_callback,
            on_error_callback=on_error_callback,
        )
    elif connection_type == "pyodide-socketio":
        if logger:
            logger.info("Using pyodide-socketio connection for imjoy-rpc")
        from .connection.pyodide_connection import PyodideConnectionManager

        manager = PyodideConnectionManager(_rpc_context)
        _rpc_context.api = dotdict(
            init=manager.init,
            export=manager.set_interface,
            registerCodec=manager.register_codec,
            register_codec=manager.register_codec,
        )

        def resolve(plugin_id):
            manager.start(
                plugin_id,
                on_ready_callback=on_ready_callback,
                on_error_callback=on_error_callback,
            )

        def reject(error):
            if on_error_callback:
                on_error_callback(error)

        setup_js_socketio(_rpc_context.default_config, resolve, reject)

    elif connection_type == "pyodide":
        if logger:
            logger.info("Using pyodide connection for imjoy-rpc")
        from .connection.pyodide_connection import PyodideConnectionManager

        manager = PyodideConnectionManager(_rpc_context)
        _rpc_context.api = dotdict(
            init=manager.init,
            export=manager.set_interface,
            registerCodec=manager.register_codec,
            register_codec=manager.register_codec,
        )
        manager.start(
            on_ready_callback=on_ready_callback, on_error_callback=on_error_callback
        )
    else:
        if logger:
            logger.info(
                "There is no connection set for imjoy-rpc, connection type: %s",
                connection_type,
            )


def type_of_script():
    """Return the type of the script."""
    try:
        import google.colab.output  # noqa: F401

        return "colab"
    except ImportError:
        try:
            import js  # noqa: F401
            import pyodide  # noqa: F401

            return "pyodide"
        except ImportError:
            try:
                # check if get_ipython exists without exporting it
                # from IPython import get_ipython
                ipy_str = str(type(get_ipython()))  # noqa: F821
                if "zmqshell" in ipy_str:
                    return "jupyter"
                if "terminal" in ipy_str:
                    return "ipython"
                else:
                    return "unknown"
            except NameError:
                return "unknown"


try:
    from js import eval, location

    _sync_xhr_get = eval(
        """globalThis._sync_xhr_get = function(url, start, end){
        var request = new XMLHttpRequest();
        request.open('GET', url, false);  // `false` makes the request synchronous
        if(start !== undefined || end !== undefined){
            request.setRequestHeader("range", `bytes=${start||0}-${end||0}`)
        }
        request.responseType = "arraybuffer";
        request.send(null);
        return request
    }
    """
    )

    _sync_xhr_post = eval(
        """globalThis._sync_xhr_post = function(url, rawData, type, append){
        var request = new XMLHttpRequest();
        request.open('POST', url, false);  // `false` makes the request synchronous
        var formData = new FormData();
        var file = new Blob([new Uint8Array(rawData)],{type});
        formData.append("file", file);
        if(append) formData.append("append", "1");
        request.send(formData);
        return request
    }
    """
    )
    IS_PYODIDE = True
except ImportError:
    from urllib.request import Request, urlopen

    IS_PYODIDE = False


class HTTPFile(io.IOBase):
    """A virtual file for reading content via HTTP."""

    def __init__(self, url, mode="r", encoding=None, newline=None):
        """Initialize the http file object."""
        self._url = url
        self._pos = 0
        self._size = None
        self._mode = mode
        assert mode in ["r", "rb", "w", "wb", "a", "ab"]
        self._encoding = encoding or locale.getpreferredencoding()
        self._newline = newline or os.linesep
        if "w" not in self._mode:
            # make a request so we can see the self._size
            self._request_range(0, 0)
            assert self._size is not None
        self._chunk = 1024
        self._initial_request = True

    def tell(self):
        """Tell the position of the pointer."""
        return self._pos

    def write(self, content):
        """Write content to file."""
        if "a" not in self._mode and "w" not in self._mode:
            raise Exception(f"write is not supported with mode {self._mode}")

        if "b" in self._mode:
            self._upload(content)
        else:
            self._upload(content.encode(self._encoding))

    def seekable(self):
        """Whether the file is seekable."""
        return "w" not in self._mode

    def read(self, length=-1):
        """Read the file from the current pointer position."""
        if self._pos >= self._size:
            return ""  # EOF
        if length < 0:
            end = self._size + length
        else:
            end = self._pos + length - 1
        if end >= self._size:
            end = self._size - 1
        result = self._request_range(self._pos, end)
        self._pos += len(result)
        if self._mode == "r":
            return result.decode(self._encoding)
        return result

    def readline(self, size=-1):
        """Read a line."""
        if self._mode == "r":
            terminator = self._newline
            result = ""
        else:
            terminator = b"\n"
            result = b""
        while True:
            ret = self.read(self._chunk)
            if ret == "":
                break
            if terminator in ret:
                used = ret.split(terminator)[0] + terminator
                unused = ret[len(used) :]
                # rollback
                self._pos -= len(unused)
                result += used
                break
            result += ret

            if size is not None and size > 0:
                if len(result) > size:
                    return result[:size]

        if not result:
            return ""
        return result

    def readlines(self, hint=-1):
        """Read all the lines."""
        if hint is None or hint < 0:
            hint = None

        lines = []
        while True:
            line = self.readline()
            if line == "":
                break
            else:
                lines.append(line)
            if hint and len(lines) >= hint:
                break
        return lines

    def seek(self, offset):
        """Set the pointer position."""
        self._pos = offset
        if self._size is not None:
            if self._pos >= self._size:
                self._pos = self._size - 1

    def _upload(self, content):
        if IS_PYODIDE:
            if self._initial_request and "a" not in self._mode:
                append = False
            else:
                append = True

            req = _sync_xhr_post(self._url, content, "application/octet-stream", append)
            if req.status != 200:
                raise Exception(f"Failed to write: {req.response}, {req.status}")

            if self._initial_request:
                self._initial_request = False
        else:
            raise NotImplementedError

    def _request_range(self, start, end):
        assert start <= end
        if IS_PYODIDE:
            req = _sync_xhr_get(self._url, start, end)
            if req.status in [200, 206]:
                result = req.response.to_py().tobytes()
                crange = req.getResponseHeader("Content-Range")
                if crange:
                    self._size = int(crange.split("/")[1])
            else:
                raise Exception(f"Failed to fetch: {req.status}")
        else:
            req = Request(self._url)
            req.add_header("range", f"bytes={start}-{end}")
            response = urlopen(req)
            if response.getcode() in [200, 206]:
                crange = response.info().getheader("Content-Range")
                if crange:
                    self._size = int(crange.split("/")[1])
                result = response.read()
            else:
                raise Exception(f"Failed to fetch: {response.getcode()}")
        return result

    def close(self):
        """Close the file."""
        self._pos = 0


def open_elfinder(path, mode="r", encoding=None, newline=None):
    """Open an HTTPFile from elFinder."""
    if not path.startswith("http"):
        url = location.origin + "/fs" + path
    else:
        url = path
    return HTTPFile(url, mode=mode, encoding=encoding, newline=newline)



import asyncio
import contextvars
import logging
import sys
import traceback
import uuid

import js
import pyodide

from imjoy_rpc.core_connection import decode_msgpack, send_as_msgpack
from imjoy_rpc.rpc import RPC
from imjoy_rpc.utils import MessageEmitter, dotdict

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("Pyodide Connection")

connection_id = contextvars.ContextVar("connection_id")
CHUNK_SIZE = 1024 * 512

# TODO: this is too high, we need to find a better approach
# see here: https://github.com/iodide-project/pyodide/issues/917#issuecomment-751307819
sys.setrecursionlimit(1500)


class PyodideConnectionManager:
    def __init__(self, rpc_context):
        self.default_config = rpc_context.default_config
        self.clients = {}
        self.interface = None
        self.rpc_context = rpc_context
        self._codecs = {}
        self.rpc_id = "pyodide_rpc"
        self.default_config["allow_execution"] = True

        loop = asyncio.get_event_loop()
        # This will not block, because we used setTimeout to execute it
        loop.run_forever()

    def get_ident(self):
        return connection_id.get(default=None)

    def set_interface(self, interface, config=None):
        config = config or self.default_config
        config = dotdict(config)
        config.id = self.rpc_id
        config.name = config.name or config.id
        config.allow_execution = config.allow_execution or False
        config.version = config.version or "0.1.0"
        config.api_version = config.api_version or "0.2.3"
        config.description = config.description or "[TODO: add description]"

        self.default_config = config
        self.interface = interface
        futures = []
        for k in self.clients:
            fut = self.clients[k].rpc.set_interface(interface, self.default_config)
            futures.append(fut)
        return asyncio.gather(*futures)

    def register_codec(self, config):
        assert "name" in config
        assert "encoder" in config or "decoder" in config
        if "type" in config:
            for tp in list(self._codecs.keys()):
                codec = self._codecs[tp]
                if codec.type == config["type"] or tp == config["name"]:
                    logger.info("Removing duplicated codec: " + tp)
                    del self._codecs[tp]

        self._codecs[config["name"]] = dotdict(config)

    def start(self, plugin_id, on_ready_callback=None, on_error_callback=None):
        try:
            self._create_new_connection(plugin_id, on_ready_callback, on_error_callback)
        except Exception as ex:
            if on_error_callback:
                on_error_callback(ex)
            raise ex

    def init(self, config=None):
        # register a minimal plugin api
        def setup():
            pass

        self.set_interface({"setup": setup}, config)

    def _create_new_connection(self, plugin_id, on_ready_callback, on_error_callback):
        client_id = str(uuid.uuid4())
        connection_id.set(client_id)
        connection = PyodideConnection(self.default_config, plugin_id)

        def initialize(data):
            self.clients[client_id] = dotdict()
            config = self.default_config.copy()
            cfg = data["config"]
            if cfg.get("credential_required") is not None:
                result = config.verify_credential(cfg["credential"])
                cfg["auth"] = result["auth"]
            cfg["id"] = self.rpc_id
            rpc = RPC(connection, self.rpc_context, config=cfg, codecs=self._codecs)
            rpc.set_interface(self.interface)
            rpc.init()

            def patch_api(_):
                api = rpc.get_remote() or dotdict()
                api.init = self.init
                api.export = self.set_interface
                api.dispose = rpc.disconnect
                api.registerCodec = self.register_codec
                api.disposeObject = rpc.dispose_object
                api.register_codec = self.register_codec
                api.dispose_object = rpc.dispose_object
                api._rpc = rpc

            rpc.on("remoteReady", patch_api)

            if on_ready_callback:

                def ready(_):
                    on_ready_callback(rpc.get_remote())

                rpc.once("interfaceSetAsRemote", ready)
            if on_error_callback:
                rpc.once("disconnected", on_error_callback)
                rpc.on("error", on_error_callback)

            self.clients[client_id].rpc = rpc

        if on_error_callback:
            connection.once("disconnected", on_error_callback)
            connection.once("error", on_error_callback)
        connection.once("initialize", initialize)
        connection.emit(
            {
                "type": "imjoyRPCReady",
                "config": dict(self.default_config),
                "peer_id": connection.peer_id,
            }
        )


def wrap_promise(promise):
    loop = asyncio.get_event_loop()
    fut = loop.create_future()
    promise.then(fut.set_result).catch(fut.set_exception)
    return fut


def install_requirements(requirements, resolve, reject):
    import micropip

    promises = []
    for r in requirements:
        p = micropip.install(r)
        promises.append(p)
    js.Promise.all(promises).then(resolve).catch(reject)


# This script template is a temporary workaround for the recursion error
# see https://github.com/iodide-project/pyodide/issues/951
script_template = """
try{
    pyodide.runPython(`%s`);
} catch(e) {
    if(e instanceof RangeError){
        console.log('Trying again due to recursion error...')
        pyodide.runPython(`%s`);
    }
    else{
        throw e
    }
}
"""


class PyodideConnection(MessageEmitter):
    def __init__(self, config, plugin_id):
        self.config = dotdict(config or {})
        super().__init__(logger)
        self.channel = self.config.get("channel") or "imjoy_rpc"
        self._event_handlers = {}
        self.peer_id = str(uuid.uuid4())
        self.debug = True
        self._send = js.sendMessage
        self.accept_encoding = []
        self.plugin_id = plugin_id

        def msg_cb(msg):
            data = msg.to_py()
            data = decode_msgpack(data, self._chunk_store)
            if data is None:
                return

            if data.get("peer_id") == self.peer_id or data.get("type") == "initialize":
                if data.get("type") == "initialize":
                    self.accept_encoding = data.get("accept_encoding", [])
                if "type" in data:
                    self._fire(data["type"], data)
            else:
                logger.warn(
                    "connection peer id mismatch {} != {}".format(
                        data.get("peer_id"), self.peer_id
                    )
                )

        js.setMessageCallback(pyodide.create_proxy(msg_cb))

    def execute(self, data):
        try:
            t = data["code"]["type"]
            if t == "script":
                content = data["code"]["content"]
                js.eval(script_template % (content, content))
                self.emit({"type": "executed"})
            elif t == "requirements":

                def success(result):
                    self.emit({"type": "executed"})

                def fail(error):
                    self.emit({"type": "executed", "error": str(error)})

                install_requirements(data["code"]["requirements"], success, fail)
            else:
                raise Exception("unsupported type")
        except Exception as e:
            traceback_error = traceback.format_exc()
            logger.error("error during execution: %s", traceback_error)
            self.emit({"type": "executed", "error": traceback_error})

    def connect(self):
        pass

    def disconnect(self):
        pass

    def emit(self, msg):
        msg["plugin_id"] = self.plugin_id
        if (
            msg.get("type") in ["initialized", "imjoyRPCReady"]
            or "msgpack" not in self.accept_encoding
        ):
            # Notify the server that the plugin supports msgpack decoding
            if msg.get("type") == "initialized":
                msg["accept_encoding"] = ["msgpack", "gzip"]
            asyncio.ensure_future(self._send(msg))
        else:
            send_as_msgpack(msg, self._send, self.accept_encoding)
